@@ -1,5 +1,5 @@
 /* Supporting functions for resolving DATA statement.
-   Copyright (C) 2002, 2003, 2004, 2005, 2007
+   Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008
    Free Software Foundation, Inc.
    Contributed by Lifang Zeng <zlf605@hotmail.com>
 
@@ -21,31 +21,32 @@ along with GCC; see the file COPYING3.  If not see
 
 
 /* Notes for DATA statement implementation:
-                                                                               
+									       
    We first assign initial value to each symbol by gfc_assign_data_value
-   during resolveing DATA statement. Refer to check_data_variable and
+   during resolving DATA statement. Refer to check_data_variable and
    traverse_data_list in resolve.c.
-                                                                               
+									       
    The complexity exists in the handling of array section, implied do
    and array of struct appeared in DATA statement.
-                                                                               
+									       
    We call gfc_conv_structure, gfc_con_array_array_initializer,
    etc., to convert the initial value. Refer to trans-expr.c and
    trans-array.c.  */
 
 #include "config.h"
 #include "gfortran.h"
+#include "data.h"
 
 static void formalize_init_expr (gfc_expr *);
 
 /* Calculate the array element offset.  */
 
 static void
-get_array_index (gfc_array_ref * ar, mpz_t * offset)
+get_array_index (gfc_array_ref *ar, mpz_t *offset)
 {
   gfc_expr *e;
   int i;
-  try re;
+  gfc_try re;
   mpz_t delta;
   mpz_t tmp;
 
@@ -60,14 +61,15 @@ get_array_index (gfc_array_ref * ar, mpz_t * offset)
       if ((gfc_is_constant_expr (ar->as->lower[i]) == 0)
 	  || (gfc_is_constant_expr (ar->as->upper[i]) == 0)
 	  || (gfc_is_constant_expr (e) == 0))
-	gfc_error ("non-constant array in DATA statement %L.", &ar->where);        
+	gfc_error ("non-constant array in DATA statement %L", &ar->where);
+
       mpz_set (tmp, e->value.integer);
       mpz_sub (tmp, tmp, ar->as->lower[i]->value.integer);
       mpz_mul (tmp, tmp, delta);
       mpz_add (*offset, tmp, *offset);
 
       mpz_sub (tmp, ar->as->upper[i]->value.integer,
-      ar->as->lower[i]->value.integer);
+	       ar->as->lower[i]->value.integer);
       mpz_add_ui (tmp, tmp, 1);
       mpz_mul (delta, tmp, delta);
     }
@@ -79,41 +81,49 @@ get_array_index (gfc_array_ref * ar, mpz_t * offset)
 /* Find if there is a constructor which offset is equal to OFFSET.  */
 
 static gfc_constructor *
-find_con_by_offset (mpz_t offset, gfc_constructor *con)
+find_con_by_offset (splay_tree spt, mpz_t offset)
 {
   mpz_t tmp;
   gfc_constructor *ret = NULL;
+  gfc_constructor *con;
+  splay_tree_node sptn;
 
+  /* The complexity is due to needing quick access to the linked list of
+     constructors.  Both a linked list and a splay tree are used, and both
+     are kept up to date if they are array elements (which is the only time
+     that a specific constructor has to be found).  */  
+
+  gcc_assert (spt != NULL);
   mpz_init (tmp);
 
-  for (; con; con = con->next)
+  sptn = splay_tree_lookup (spt, (splay_tree_key) mpz_get_si (offset));
+
+  if (sptn)
+    ret = (gfc_constructor*) sptn->value;  
+  else
     {
-      int cmp = mpz_cmp (offset, con->n.offset);
-
-      /* We retain a sorted list, so if we're too large, we're done.  */
-      if (cmp < 0)
-	break;
-
-      /* Yaye for exact matches.  */
-      if (cmp == 0)
-	{
-          ret = con;
-	  break;
-	}
-
-      /* If the constructor element is a range, match any element.  */
-      if (mpz_cmp_ui (con->repeat, 1) > 0)
-	{
-	  mpz_add (tmp, con->n.offset, con->repeat);
-	  if (mpz_cmp (offset, tmp) < 0)
-	    {
-	      ret = con;
-	      break;
-	    }
-	}
+       /* Need to check and see if we match a range, so we will pull
+	  the next lowest index and see if the range matches.  */
+       sptn = splay_tree_predecessor (spt,
+				      (splay_tree_key) mpz_get_si (offset));
+       if (sptn)
+	 {
+	    con = (gfc_constructor*) sptn->value;
+	    if (mpz_cmp_ui (con->repeat, 1) > 0)
+	      {
+		 mpz_init (tmp);
+		 mpz_add (tmp, con->n.offset, con->repeat);
+		 if (mpz_cmp (offset, tmp) < 0)
+		   ret = con;
+		 mpz_clear (tmp);
+	      }
+	    else 
+	      ret = NULL; /* The range did not match.  */
+	 }
+      else
+	ret = NULL; /* No pred, so no match.  */
     }
 
-  mpz_clear (tmp);
   return ret;
 }
 
@@ -126,7 +136,7 @@ find_con_by_component (gfc_component *com, gfc_constructor *con)
   for (; con; con = con->next)
     {
       if (com == con->n.component)
-        return con;
+	return con;
     }
   return NULL;
 }
@@ -134,17 +144,15 @@ find_con_by_component (gfc_component *com, gfc_constructor *con)
 
 /* Create a character type initialization expression from RVALUE.
    TS [and REF] describe [the substring of] the variable being initialized.
-   INIT is thh existing initializer, not NULL.  Initialization is performed
+   INIT is the existing initializer, not NULL.  Initialization is performed
    according to normal assignment rules.  */
 
 static gfc_expr *
-create_character_intializer (gfc_expr * init, gfc_typespec * ts,
-			     gfc_ref * ref, gfc_expr * rvalue)
+create_character_intializer (gfc_expr *init, gfc_typespec *ts,
+			     gfc_ref *ref, gfc_expr *rvalue)
 {
-  int len;
-  int start;
-  int end;
-  char *dest;
+  int len, start, end;
+  gfc_char_t *dest;
 	    
   gfc_extract_int (ts->cl->length, &len);
 
@@ -155,13 +163,13 @@ create_character_intializer (gfc_expr * init, gfc_typespec * ts,
       init->expr_type = EXPR_CONSTANT;
       init->ts = *ts;
       
-      dest = gfc_getmem (len + 1);
+      dest = gfc_get_wide_string (len + 1);
       dest[len] = '\0';
       init->value.character.length = len;
       init->value.character.string = dest;
       /* Blank the string if we're only setting a substring.  */
       if (ref != NULL)
-	memset (dest, ' ', len);
+	gfc_wide_memset (dest, ' ', len);
     }
   else
     dest = init->value.character.string;
@@ -173,14 +181,14 @@ create_character_intializer (gfc_expr * init, gfc_typespec * ts,
       gcc_assert (ref->type == REF_SUBSTRING);
 
       /* Only set a substring of the destination.  Fortran substring bounds
-         are one-based [start, end], we want zero based [start, end).  */
+	 are one-based [start, end], we want zero based [start, end).  */
       start_expr = gfc_copy_expr (ref->u.ss.start);
       end_expr = gfc_copy_expr (ref->u.ss.end);
 
       if ((gfc_simplify_expr (start_expr, 1) == FAILURE)
-	     || (gfc_simplify_expr (end_expr, 1)) == FAILURE)
+	  || (gfc_simplify_expr (end_expr, 1)) == FAILURE)
 	{
-	  gfc_error ("failure to simplify substring reference in DATA"
+	  gfc_error ("failure to simplify substring reference in DATA "
 		     "statement at %L", &ref->u.ss.start->where);
 	  return NULL;
 	}
@@ -197,7 +205,11 @@ create_character_intializer (gfc_expr * init, gfc_typespec * ts,
     }
 
   /* Copy the initial value.  */
-  len = rvalue->value.character.length;
+  if (rvalue->ts.type == BT_HOLLERITH)
+    len = rvalue->representation.length;
+  else
+    len = rvalue->value.character.length;
+
   if (len > end - start)
     {
       len = end - start;
@@ -205,33 +217,50 @@ create_character_intializer (gfc_expr * init, gfc_typespec * ts,
 		       "at %L", &rvalue->where);
     }
 
-  memcpy (&dest[start], rvalue->value.character.string, len);
+  if (rvalue->ts.type == BT_HOLLERITH)
+    {
+      int i;
+      for (i = 0; i < len; i++)
+	dest[start+i] = rvalue->representation.string[i];
+    }
+  else
+    memcpy (&dest[start], rvalue->value.character.string,
+	    len * sizeof (gfc_char_t));
 
   /* Pad with spaces.  Substrings will already be blanked.  */
   if (len < end - start && ref == NULL)
-    memset (&dest[start + len], ' ', end - (start + len));
+    gfc_wide_memset (&dest[start + len], ' ', end - (start + len));
 
   if (rvalue->ts.type == BT_HOLLERITH)
-    init->from_H = 1;
+    {
+      init->representation.length = init->value.character.length;
+      init->representation.string
+	= gfc_widechar_to_char (init->value.character.string,
+				init->value.character.length);
+    }
 
   return init;
 }
 
+
 /* Assign the initial value RVALUE to  LVALUE's symbol->value. If the
    LVALUE already has an initialization, we extend this, otherwise we
    create a new one.  */
 
-void
-gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
+gfc_try
+gfc_assign_data_value (gfc_expr *lvalue, gfc_expr *rvalue, mpz_t index)
 {
   gfc_ref *ref;
   gfc_expr *init;
   gfc_expr *expr;
   gfc_constructor *con;
   gfc_constructor *last_con;
+  gfc_constructor *pred;
   gfc_symbol *symbol;
   gfc_typespec *last_ts;
   mpz_t offset;
+  splay_tree spt;
+  splay_tree_node sptn;
 
   symbol = lvalue->symtree->n.sym;
   init = symbol->value;
@@ -251,7 +280,7 @@ gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
 	}
 
       /* Use the existing initializer expression if it exists.  Otherwise
-         create a new one.  */
+	 create a new one.  */
       if (init == NULL)
 	expr = gfc_get_expr ();
       else
@@ -261,6 +290,14 @@ gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
       switch (ref->type)
 	{
 	case REF_ARRAY:
+	  if (init && expr->expr_type != EXPR_ARRAY)
+	    {
+	      gfc_error ("'%s' at %L already is initialized at %L",
+			 lvalue->symtree->n.sym->name, &lvalue->where,
+			 &init->where);
+	      return FAILURE;
+	    }
+
 	  if (init == NULL)
 	    {
 	      /* The element typespec will be the same as the array
@@ -270,24 +307,69 @@ gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
 	      expr->expr_type = EXPR_ARRAY;
 	      expr->rank = ref->u.ar.as->rank;
 	    }
-	  else
-	    gcc_assert (expr->expr_type == EXPR_ARRAY);
 
 	  if (ref->u.ar.type == AR_ELEMENT)
 	    get_array_index (&ref->u.ar, &offset);
 	  else
 	    mpz_set (offset, index);
 
-	  /* Find the same element in the existing constructor.  */
-	  con = expr->value.constructor;
-	  con = find_con_by_offset (offset, con);
+	  /* Check the bounds.  */
+	  if (mpz_cmp_si (offset, 0) < 0)
+	    {
+	      gfc_error ("Data element below array lower bound at %L",
+			 &lvalue->where);
+	      return FAILURE;
+	    }
+	  else
+	    {
+	      mpz_t size;
+	      if (spec_size (ref->u.ar.as, &size) == SUCCESS)
+		{
+		  if (mpz_cmp (offset, size) >= 0)
+		  {
+		    mpz_clear (size);
+		    gfc_error ("Data element above array upper bound at %L",
+			       &lvalue->where);
+		    return FAILURE;
+		  }
+		  mpz_clear (size);
+		}
+	    }
+
+	  /* Splay tree containing offset and gfc_constructor.  */
+	  spt = expr->con_by_offset;
+
+	  if (spt == NULL)
+	    {
+	       spt = splay_tree_new (splay_tree_compare_ints, NULL, NULL);
+	       expr->con_by_offset = spt; 
+	       con = NULL;
+	    }
+	 else
+	  con = find_con_by_offset (spt, offset);
 
 	  if (con == NULL)
 	    {
+	      splay_tree_key j;
+
 	      /* Create a new constructor.  */
 	      con = gfc_get_constructor ();
 	      mpz_set (con->n.offset, offset);
-	      gfc_insert_constructor (expr, con);
+	      j = (splay_tree_key) mpz_get_si (offset);
+	      sptn = splay_tree_insert (spt, j, (splay_tree_value) con);
+	      /* Fix up the linked list.  */
+	      sptn = splay_tree_predecessor (spt, j);
+	      if (sptn == NULL)
+		{  /* Insert at the head.  */
+		   con->next = expr->value.constructor;
+		   expr->value.constructor = con;
+		}
+	      else
+		{  /* Insert in the chain.  */
+		   pred = (gfc_constructor*) sptn->value;
+		   con->next = pred->next;
+		   pred->next = con;
+		}
 	    }
 	  break;
 
@@ -341,19 +423,14 @@ gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
 	 provokes a warning from other compilers.  */
       if (init != NULL)
 	{
-	  /* Order in which the expressions arrive here depends on whether they
-	     are from data statements or F95 style declarations. Therefore,
-	     check which is the most recent.  */
-#ifdef USE_MAPPED_LOCATION
+	  /* Order in which the expressions arrive here depends on whether
+	     they are from data statements or F95 style declarations.
+	     Therefore, check which is the most recent.  */
 	  expr = (LOCATION_LINE (init->where.lb->location)
 		  > LOCATION_LINE (rvalue->where.lb->location))
-	    ? init : rvalue;
-#else
-	  expr = (init->where.lb->linenum > rvalue->where.lb->linenum) ?
-		    init : rvalue;
-#endif
+	       ? init : rvalue;
 	  gfc_notify_std (GFC_STD_GNU, "Extension: re-initialization "
-			  "of '%s' at %L",  symbol->name, &expr->where);
+			  "of '%s' at %L", symbol->name, &expr->where);
 	}
 
       expr = gfc_copy_expr (rvalue);
@@ -365,22 +442,28 @@ gfc_assign_data_value (gfc_expr * lvalue, gfc_expr * rvalue, mpz_t index)
     symbol->value = expr;
   else
     last_con->expr = expr;
+
+  return SUCCESS;
 }
 
+
 /* Similarly, but initialize REPEAT consecutive values in LVALUE the same
    value in RVALUE.  For the nonce, LVALUE must refer to a full array, not
    an array section.  */
 
 void
-gfc_assign_data_value_range (gfc_expr * lvalue, gfc_expr * rvalue,
+gfc_assign_data_value_range (gfc_expr *lvalue, gfc_expr *rvalue,
 			     mpz_t index, mpz_t repeat)
 {
   gfc_ref *ref;
   gfc_expr *init, *expr;
   gfc_constructor *con, *last_con;
+  gfc_constructor *pred;
   gfc_symbol *symbol;
   gfc_typespec *last_ts;
   mpz_t offset;
+  splay_tree spt;
+  splay_tree_node sptn;
 
   symbol = lvalue->symtree->n.sym;
   init = symbol->value;
@@ -434,17 +517,43 @@ gfc_assign_data_value_range (gfc_expr * lvalue, gfc_expr * rvalue,
 	    }
 
 	  /* Find the same element in the existing constructor.  */
-	  con = expr->value.constructor;
-	  con = find_con_by_offset (offset, con);
 
-	  /* Create a new constructor.  */
+	  /* Splay tree containing offset and gfc_constructor.  */
+	  spt = expr->con_by_offset;
+
+	  if (spt == NULL)
+	    {
+	       spt = splay_tree_new (splay_tree_compare_ints, NULL, NULL);
+	       expr->con_by_offset = spt;
+	       con = NULL;
+	    }
+	  else 
+	    con = find_con_by_offset (spt, offset);
+
 	  if (con == NULL)
 	    {
+	      splay_tree_key j;
+	      /* Create a new constructor.  */
 	      con = gfc_get_constructor ();
 	      mpz_set (con->n.offset, offset);
+	      j = (splay_tree_key) mpz_get_si (offset);
+	  
 	      if (ref->next == NULL)
 		mpz_set (con->repeat, repeat);
-	      gfc_insert_constructor (expr, con);
+	      sptn = splay_tree_insert (spt, j, (splay_tree_value) con);
+	      /* Fix up the linked list.  */
+	      sptn = splay_tree_predecessor (spt, j);
+	      if (sptn == NULL)
+		{  /* Insert at the head.  */
+		   con->next = expr->value.constructor;
+		   expr->value.constructor = con;
+		}
+	      else
+		{  /* Insert in the chain.  */
+		   pred = (gfc_constructor*) sptn->value;
+		   con->next = pred->next;
+		   pred->next = con;
+		}
 	    }
 	  else
 	    gcc_assert (ref->next != NULL);
@@ -552,10 +661,9 @@ gfc_advance_section (mpz_t *section_index, gfc_array_ref *ar,
       else
 	cmp = mpz_cmp (section_index[i], ar->as->upper[i]->value.integer);
 
-      if ((cmp > 0 && forwards)
-	  || (cmp < 0 && ! forwards))
+      if ((cmp > 0 && forwards) || (cmp < 0 && !forwards))
 	{
-          /* Reset index to start, then loop to advance the next index.  */
+	  /* Reset index to start, then loop to advance the next index.  */
 	  if (ar->start[i])
 	    mpz_set (section_index[i], ar->start[i]->value.integer);
 	  else
@@ -575,7 +683,7 @@ gfc_advance_section (mpz_t *section_index, gfc_array_ref *ar,
       mpz_add (*offset_ret, tmp, *offset_ret);
 
       mpz_sub (tmp, ar->as->upper[i]->value.integer, 
-               ar->as->lower[i]->value.integer);
+	       ar->as->lower[i]->value.integer);
       mpz_add_ui (tmp, tmp, 1);
       mpz_mul (delta, tmp, delta);
     }
@@ -588,7 +696,7 @@ gfc_advance_section (mpz_t *section_index, gfc_array_ref *ar,
    order.  Also insert NULL entries if necessary.  */
 
 static void
-formalize_structure_cons (gfc_expr * expr)
+formalize_structure_cons (gfc_expr *expr)
 {
   gfc_constructor *head;
   gfc_constructor *tail;
@@ -600,7 +708,7 @@ formalize_structure_cons (gfc_expr * expr)
   c = expr->value.constructor;
 
   /* Constructor is already formalized.  */
-  if (c->n.component == NULL)
+  if (!c || c->n.component == NULL)
     return;
 
   head = tail = NULL;
@@ -646,11 +754,11 @@ formalize_structure_cons (gfc_expr * expr)
 }
 
 
-/* Make sure an initialization expression is in normalized form.  Ie. all
+/* Make sure an initialization expression is in normalized form, i.e., all
    elements of the constructors are in the correct order.  */
 
 static void
-formalize_init_expr (gfc_expr * expr)
+formalize_init_expr (gfc_expr *expr)
 {
   expr_t type;
   gfc_constructor *c;
@@ -729,7 +837,7 @@ gfc_get_section_index (gfc_array_ref *ar, mpz_t *section_index, mpz_t *offset)
 	}
 
       mpz_sub (tmp, ar->as->upper[i]->value.integer, 
-               ar->as->lower[i]->value.integer);
+	       ar->as->lower[i]->value.integer);
       mpz_add_ui (tmp, tmp, 1);
       mpz_mul (delta, tmp, delta);
     }
